#!/usr/bin/perl

# This script takes the data from a msblast result file and call clustalw2 on them
# Written by Alexandre Burel, IPHC LSMBO (March 2012)

use strict;
use warnings;

# constant declarations
use constant SCORE => 30;
use constant AMINO_A_WT => 89.09;
use constant AMINO_C_WT => 121.15;
use constant AMINO_D_WT => 133.1;
use constant AMINO_E_WT => 147.13;
use constant AMINO_F_WT => 165.19;
use constant AMINO_G_WT => 75.07;
use constant AMINO_H_WT => 155.16;
use constant AMINO_I_WT => 131.18;
use constant AMINO_K_WT => 146.19;
use constant AMINO_L_WT => 131.18;
use constant AMINO_M_WT => 149.22;
use constant AMINO_N_WT => 132.12;
use constant AMINO_P_WT => 115.13;
use constant AMINO_Q_WT => 146.15;
use constant AMINO_R_WT => 174.21;
use constant AMINO_S_WT => 105.09;
use constant AMINO_T_WT => 119.12;
use constant AMINO_V_WT => 117.15;
use constant AMINO_W_WT => 204.22;
use constant AMINO_Y_WT => 181.19;
use constant WATER => 18.015;

# main arguments
my $input_file = $ARGV[0];
my $mgf_file = $ARGV[1];
my $pepnovo_results = $ARGV[2];
my $xls_output_file = $ARGV[3];
my $mgf_output_file = $ARGV[4];

# checking arguments
if(! -e $input_file)
{
	print "\nFile $input_file does not exist\n";
	exit 1;
}
if(! -e $mgf_file)
{
	print "\nFile  $mgf_file does not exist\n\n";
	exit 1;
}
if(! -e $pepnovo_results)
{
	print "\nFile  $pepnovo_results does not exist\n\n";
	exit 1;
}

# the list of amino acids weights (used to process the molecular weight)
my %AMINO_WEIGHTS = (
	'A' => [AMINO_A_WT, AMINO_A_WT], # Alanine
	'B' => [AMINO_N_WT, AMINO_D_WT], # Aspartic Acid, Asparagine
	'C' => [AMINO_C_WT, AMINO_C_WT], # Cystine
	'D' => [AMINO_D_WT, AMINO_D_WT], # Aspartic Acid
	'E' => [AMINO_E_WT, AMINO_E_WT], # Glutamic Acid
	'F' => [AMINO_F_WT, AMINO_F_WT], # Phenylalanine
	'G' => [AMINO_G_WT, AMINO_G_WT], # Glycine
	'H' => [AMINO_H_WT, AMINO_H_WT], # Histidine
	'I' => [AMINO_I_WT, AMINO_I_WT], # Isoleucine
	'K' => [AMINO_K_WT, AMINO_K_WT], # Lysine
	'L' => [AMINO_L_WT, AMINO_L_WT], # Leucine
	'M' => [AMINO_M_WT, AMINO_M_WT], # Methionine
	'N' => [AMINO_N_WT, AMINO_N_WT], # Asparagine
	'P' => [AMINO_P_WT, AMINO_P_WT], # Proline
	'Q' => [AMINO_Q_WT, AMINO_Q_WT], # Glutamine
	'R' => [AMINO_R_WT, AMINO_R_WT], # Arginine
	'S' => [AMINO_S_WT, AMINO_S_WT], # Serine
	'T' => [AMINO_T_WT, AMINO_T_WT], # Threonine
	'V' => [AMINO_V_WT, AMINO_V_WT], # Valine
	'W' => [AMINO_W_WT, AMINO_W_WT], # Tryptophan
	'X' => [AMINO_G_WT, AMINO_W_WT], # Unknown
	'Y' => [AMINO_Y_WT, AMINO_Y_WT], # Tyrosine
	'Z' => [AMINO_Q_WT, AMINO_E_WT]  # Glutamic Acid, Glutamine
);

# global arrays for the excel file
my @msblastQueries;
my @spectrumMatches;

# searching the list of spectrum titles in the mgf file
open(MGF, "$mgf_file") || die("Can't open $mgf_file as MGF");
my @spectrum_titles;
while(my $line = <MGF>)
{
	$line =~ s/\r?\n?$//;
	push(@spectrum_titles, $1) if($line =~ m/TITLE=(.*)/);
}
close(MGF);

# reading the protein grouping output file
my @proteins;
#my %protein;
my $file = ""; my $acc = ""; my $desc = ""; my $org = ""; my $db = ""; my $score = ""; my $seq = ""; my @peptides;
open(INPUT, "$input_file") || die("Can't open $input_file as INPUT");
while(my $line = <INPUT>)
{
	$line =~ s/\r?\n?$//;
	if($line =~ m/^BEGIN$/)
	{
		# initialize the hash
		$file = ""; $acc = ""; $desc = ""; $org = ""; $db = ""; $score = ""; $seq = ""; @peptides = ();
	}
	$file = $1 if($line =~ m/^FILE=(.*)/);
	$acc = $1 if($line =~ m/^ACC=(.*)/);
	$desc = $1 if($line =~ m/^DESC=(.*)/);
	$org = $1 if($line =~ m/^ORG=(.*)/);
	$db = $1 if($line =~ m/^DB=(.*)/);
	$score = $1 if($line =~ m/^SCORE=(.*)/);
	$seq = $1 if($line =~ m/^SEQUENCE=(.*)/);
	push(@peptides, $1) if($line =~ m/^PEPTIDE=(.*)/);
	if($line =~ m/^END$/)
	{
		# record the hash
		my @tmp = @peptides;
		my %protein = ("file", $file, "accession", $acc, "description", $desc, "organism", $org, "database", $db, "score", $score, "sequence", $seq, "peptides", \@tmp);
		push(@proteins, \%protein);
	}
}
close(INPUT);

# write the excel output file
my $xls_separator = "\t";
open(XLS, ">$xls_output_file") || die("Can't open $xls_output_file as XLS");
my @headers = (
	"MS/MS Sample name",
	"Protein AN",
	"Protein name",
	"Organism",
	"Database",
	"Protein Molecular Weight(Da)",
	"Number of peptides in protein",
	"MSBlast query",
	"MSBlast Score",
	"Spectrum title"
);
print XLS join($xls_separator, @headers)."\n";
if(scalar(@proteins) == 0)
{
	print XLS REMOVE_EXTENSION($mgf_file)."\n";
}
else
{
	my @titles_of_matching_spectra;
	foreach my $pt (@proteins)
	{
		my %protein = %{$pt};
		my $nbPeptides = getExtendedInfoAboutThePeptides($pepnovo_results, \@spectrum_titles, @{$protein{peptides}});
		for(my $i = 0; $i < $nbPeptides; $i++)
		{
			my @line = (
				REMOVE_EXTENSION($protein{file}), 
				$protein{accession}, 
				$protein{description}, 
				$protein{organism}, 
				REMOVE_EXTENSION($protein{database}), 
				molecular_weight($protein{sequence}, \%AMINO_WEIGHTS), 
				$nbPeptides,
				$msblastQueries[$i],
				$protein{score},
				$spectrumMatches[$i]
			);
			print XLS join($xls_separator, @line)."\n";
		}
		# adding spectrum to the list of spectra that matched in pepnovo
		push(@titles_of_matching_spectra, @spectrumMatches);
	}
	# remove the duplicate titles from the array
	my %unique_titles = map { $_ => 1 } @titles_of_matching_spectra;
	# adding spectra to the $mgf_with_matching_spectra file
	extract_spectra_from_peaklist($mgf_output_file, $mgf_file, sort(keys(%unique_titles)));
}
close(XLS);

exit;

sub getExtendedInfoAboutThePeptides
{
	my ($pepnovo, $refTitles, @peptides) = @_;
	my @titles = @{$refTitles};
	# empty the arrays
	@msblastQueries = ();
	@spectrumMatches = ();
	# read the pepnovo result file and save the sequence and the spectrum ids
	my %pepnovo;
	if(open(PEP, "$pepnovo"))
	{
		while(<PEP>)
		{
			$_ =~ s/\.//g; # removing the "." characters because they aren't sent to 
			$_ =~ /^\d+\t(\d+)\t.*\-([A-Z\.]+)$/;
			$pepnovo{$2} = $1; # sequence => spectrum id
		}
		close(PEP);
	}
	# look for each peptide if it can be found in the pepnovo result
	foreach my $peptide (@peptides)
	{
		foreach my $sequence (keys(%pepnovo))
		{
			if($sequence =~ m/.*$peptide.*/) # does pepnovo sequence contains msblast sequence ?
			{
				push(@msblastQueries, $peptide);
				push(@spectrumMatches, $titles[$pepnovo{$sequence}]);
				last;
			}
		}
	}
	# the arrays must have the same size
	my $nb = 0;
	if(scalar(@msblastQueries) == scalar(@spectrumMatches))
	{
		$nb = scalar(@msblastQueries);
	}
	return $nb;
}

# search spectrum title of the given peptide sequences in the pepnovo result file
# get the spectrum title from the given title list
# $pepnovo : the pepnovo result file
# $refTitles : a reference to the spectrum title list
# @peptides : an array containing the peptide sequences to search
# return a hash containing the sequences and their full title ($peptide=$title)
#sub search_titles
#{
#	my ($pepnovo, $refTitles, @peptides) = @_;
#	my @titles = @{$refTitles};
#	my %peptides_with_title;
#	my %pepnovo;
#	open(PEP, "$pepnovo") || die("Can't open $pepnovo as PEP");
#	while(my $line = <PEP>)
#	{
#		$line =~ s/\r?\n?$//;
#		#$line =~ /^[^\t]+\t([^\t]+).*\t\-?([^\t]+)\t*$/;
#		$line =~ /^\d+\t(\d+)\t.*\-(\w+)$/;
#		$pepnovo{$2} = $1; # sequence => spectrum id
#	}
#	close(PEP);
#	foreach my $peptide (@peptides)
#	{
#		$peptides_with_title{$peptide} = "?";
#		foreach my $sequence (keys(%pepnovo))
#		{
#			if($sequence =~ m/$peptide/) # does pepnovo sequence contains msblast sequence ?
#			{
#				$peptides_with_title{$peptide} = $titles[$pepnovo{$sequence}];
#				last;
#			}
#		}
#	}
#	return %peptides_with_title;
#}

sub extract_spectra_from_peaklist
{
	my ($new_mgf_file, $mgf_file, @spectrum_titles) = @_;
	if($new_mgf_file ne "")
	{
		open(OUT, ">$new_mgf_file") || die("Can't open $new_mgf_file as OUT");
		# the file must be filled only with the given spectra
		if(scalar(@spectrum_titles) > 0)
		{
			my %titles = map { $_ => 1 } @spectrum_titles;
			open(MGF, "$mgf_file") || die("Can't open $mgf_file as MGF");
			local $/ = "END IONS";
			while(my $spectrum = <MGF>)
			{
				$spectrum =~ /TITLE=([^\n]*)/;
				my $title = $1;
				next if(!$title || $title eq ""); # avoid empty block
				$title =~ s/\r?\n?$//; # remove trailing \r if file written on windows system
				# write the spectrum if it is in the list
				print OUT $spectrum if(exists($titles{$title}));
			}
			close(MGF);
		}
		close(OUT);
	}
}

# calculates the molecular weight of an amino acid sequence
# $sequence : the amino acid sequence
# return the molecular weight
# inspired by Bioperl http://doc.bioperl.org/releases/bioperl-1.0.1/Bio/Tools/SeqStats.html#vardefs
sub molecular_weight
{
	my $sequence = $_[0];
	my %amino_weights = %{$_[1]};
	my %alphabets = (
		'protein' => [ qw(A R N D C Q E G H I L K M F P S T W X Y V B Z ) ]
	);

	my %result;
	# if the sequence contains other characters than those in the alphabet
	my $alpha_array = $alphabets{'protein'};
	my $alphabet = join ('', @$alpha_array);
	return "ERROR" if(!$sequence || $sequence =~ /[^$alphabet]/);

	my %count;
	foreach my $element (@$alpha_array)
	{
		$count{$element} = ($sequence =~ s/$element/$element/g);
	}
	my $weight_lower_bound = 0;
	my $weight_upper_bound = 0;

	foreach my $key (keys %count)
	{
		$weight_lower_bound += $count{$key} * $amino_weights{$key}->[0];
		$weight_upper_bound += $count{$key} * $amino_weights{$key}->[1];
	}
	$weight_lower_bound -= WATER * (length($sequence) - 1);
	$weight_upper_bound -= WATER * (length($sequence) - 1);
	$weight_lower_bound = sprintf("%.0f", $weight_lower_bound);
	$weight_upper_bound = sprintf("%.0f", $weight_upper_bound);
	return $weight_lower_bound." - ".$weight_upper_bound if($weight_lower_bound != $weight_upper_bound);
	return $weight_lower_bound;
}

sub REMOVE_EXTENSION
{
	my $arg = $_[0];
	$arg =~ s/(.*)\.[a-z]+$/$1/ if($arg =~ m/\./);
	return $arg;
}

