#!/usr/bin/perl

# This script takes the data from a msblast result file and call clustalw2 on them
# Written by Alexandre Burel, IPHC LSMBO (February 2012)
#
# Modified by Alexandre Burel, IPHC LSMBO (March 2012)
# adding support to non-gi entries in the database
# replaced the excel creation file by a temporary text file
# the excel file is created in another script
#
# Modified by Alexandre Burel, IPHC LSMBO (September 2012)
# corrected the extraction of the score from the input file

use strict;
use warnings;

# constant declarations
use constant SCORE => 30;

# main arguments
my $file = $ARGV[0];
my $database = $ARGV[1];
my $output_file = $ARGV[2];
my $CLUSTALW2_SCORE = SCORE;
$CLUSTALW2_SCORE = $ARGV[3] if(scalar(@ARGV) >= 4 && $ARGV[3] =~ m/^[0-9]+$/);

# checking arguments
if(! -e $file)
{
	print "\nFile $file does not exist\n";
	exit 1;
}
if(! -e $database)
{
	print "\nFile  $database does not exist\n\n";
	exit 1;
}

# %acc hash format : 
# %acc = {
#	$acc => {
#		"description" = $description, 
#		"score" = $score, 
#		"organism" = $organism, 
#		"sequence" = $sequence
#	}
# }
my %acc;	# the main accession number hash
my %scores;	# the list of the accession numbers and the corresponding score (used for clustalw2)
my @acc;	# the list of accession numbers to add to the output file



print "STEP 1 : READING THE INPUT FILE\n";
# extraction of the red accession numbers in the file
open(INPUT, "$file") || die("Can't open $file as INPUT");
# put the cursor at the beginning of the "red" accession numbers list
do {} until (<INPUT> =~ /^\<font color=\'#990000\'\>/);
# read until the end of the "red" accession number list
while(my $line = <INPUT>)
{
	$line =~ s/\r?\n?$//;
	last if($line =~ m/^\<\/font\>/);
	next if($line =~ m/REVERSED_/ || $line =~ m/REV_/);
	next if($line !~ m/\/:..\|.*[\s\|].*[0-9]+$/); # precaution if the database has not a correct format
	# line example : 
	# /:gi|99082883|gb|ABC59068.2| alpha tubulin 1 [Camellia si...    70  70
#	$line =~ /\/:(..)\|([A-Z0-9]+).*\s+([0-9]+)$/;
	$line =~ /\/:(..)\|([^\s\|]+).*[^0-9]([0-9]+)$/;
	my ($dbtype, $acc, $score) = ($1, $2, $3);
	# we consider that a accession number does not appear more than once
	# if it does, only the last accession number will be recorder
	$acc{$acc} = {};
	$acc{$acc}{score} = $score;
	$acc{$acc}{dbtype} = $dbtype;
	$scores{$acc} = $score;
}
close(INPUT);


if(scalar(keys(%acc)) > 0)
{
	print "STEP 2 : READING THE DATABASE\n";
	# open the database to complete the hash
	open(DB, "$database") || die("Can't open $database as DB");
	my $current_acc = "";
	while(my $line = <DB>)
	{
		$line =~ s/\r?\n?$//;
		#if($line =~ m/^>\/:..\|/)
		if($line =~ m/^>\/:([^\|]+)\|([^\s\|]+)[^\s]*\s+(.*)$/)
		{
			$current_acc = "";
			# line example : 
			# >/:gi|293218|gb|AAA20111.1| chlorophyll a/b binding protein [Mantoniella squamata]
			#$line =~ /^>\/:(..)\|([^\s\|]+)[^\s]*\s+(.*)$/;
			my ($dbtype, $acc, $description) = ($1, $2, $3);
			# do nothing if the accession number is not in red in $file
			next if(!exists($acc{$acc}) || $acc{$acc}{dbtype} ne $dbtype);
			# if the accession number exists but corresponds to trypsin or keratin, remove it from the hash
#			if(lc($description) =~ m/keratin/ || lc($description) =~ m/trypsin/)
#			{
#				delete $acc{$acc};
#				delete $scores{$acc} if(exists($scores{$acc}));
#				next;
#			}
			# otherwise, get the sequence and record data in the hash
			$current_acc = $acc;
			$acc{$acc}{dbtype} = $dbtype;
			$acc{$acc}{description} = $description;
			$acc{$acc}{organism} = "";
			if($description =~ m/\[.*\]/)
			{
				$description =~ /.*(\[.*\])/;
				$acc{$acc}{organism} = $1;
			}
			$acc{$acc}{sequence} = "";
		}
		elsif($current_acc ne "")
		{
			$acc{$current_acc}{sequence} .= $line;
		}
	}
	close(DB);


	if(scalar(keys(%acc)) > 0)
	{
		# at this point, we no longer need the database
		if($CLUSTALW2_SCORE == 100)
		{
			print "STEP 3 : SKIP CALLING CLUSTALW2 FOR EACH ACCESSION NUMBER (SCORE = $CLUSTALW2_SCORE)\n";
			@acc = keys %acc;
		}
		else
		{
			print "STEP 3 : CALLING CLUSTALW2 FOR EACH ACCESSION NUMBER (SCORE = $CLUSTALW2_SCORE)\n";
			my @scores = sort {$scores{$b} <=> $scores{$a}} (keys %scores);
			my %tmp_scores;
			my $nb_scores = scalar(keys %tmp_scores);
			do
			{
				my $first_acc = $scores[0];
				my $first_sequence = $acc{$first_acc}{sequence};
				my $first_dbtype = $acc{$first_acc}{dbtype};
				# this accession number will be written in the output file
				push (@acc, $first_acc);
				# comparing the first accession number to the others in descending order
				foreach my $acc (@scores)
				{
					next if($acc{$acc}{dbtype} ne "gi" && $first_acc eq $acc);
					next if($acc{$acc}{dbtype} eq "gi" && $first_acc == $acc);
					# preparing the input file for clustalw2
					my $clustalw2_input = NEWFILE();
					open(OUT, ">$clustalw2_input") || die("Can't open $clustalw2_input as OUT");
					print OUT ">".$first_dbtype."|".$first_acc."\n";
					print OUT $first_sequence."\n";
					print OUT ">".$acc{$acc}{dbtype}."|".$acc."\n";
					print OUT $acc{$acc}{sequence}."\n";
					close (OUT);
					# executing clustalw2
					my $clustalw2_output = NEWFILE();
					system ("./clustalw2_64 -infile=$clustalw2_input -align -type=protein -ktuple=1 -window=5 -score=percent -topdiags=5 -pairgap=3 -pwmatrix=gonnet -pwdnamatrix=iub -pwgapopen=10 -pwgapext=0.1 -matrix=gonnet -dnamatrix=iub -gapopen=10 -gapext=0.2 -gapdist=5 -iteration=none -numiter=1 -clustering=NJ -seqnos=on -outorder=aligned > $clustalw2_output");
					unlink($clustalw2_input);
					# reading the clustalw2 output
					open(CLUST, "$clustalw2_output") || die("Can't open $clustalw2_output as INPUT");
					# place the cursor at the begginning of the lines of interest
					do {} until (<CLUST> =~ /Sequence format is/);
					# record the accession number and sequence number
					my %clust_acc;
					while(my $line = <CLUST>)
					{
						$line =~ s/\r?\n?$//;
						last if($line =~ /Start of Pairwise alignments/);
						# line example : 
						# Sequence 1: gi|57867853    257 aa
	#					$line =~ /^.*\s(\d+):.*\|([A-Z0-9]+).*/;
						$line =~ /^.*\s(\d+):.*\|([^\s]+).*/;
						my $num_sequence = $1;
						my $clust_acc = $2;
						$clust_acc{$num_sequence} = $clust_acc;
					}
					# record the sequences
					while(my $line = <CLUST>)
					{
						$line =~ s/\r?\n?$//;
						if($line =~ /^Sequences \(/)
						{
							# line example : 
							# Sequences (1:2) Aligned. Score:  75.004
							$line =~ /.*\s+\((\d+):(\d+).*:\s+(\d+\.\d+|\d+)$/;
							my $clust_sequence_1 = $1; # first sequence, should always be 1
							my $clust_sequence_2 = $2; # second sequence, should always be 2
							my $clust_score = $3; # score of the comparison of seq1 and seq2
							if($clust_sequence_2 == 2 && $clust_score <= $CLUSTALW2_SCORE)
							{
								# if the score is less than 30, the accession number is recorder is recorded
								$tmp_scores{$clust_acc{$clust_sequence_2}} = $acc{$clust_acc{$clust_sequence_2}}{score};
							}
						}
					}
					close(CLUST);
					unlink($clustalw2_output);
				}
				# only the accession number with a clustalw2 score less than 30 are kept
				$nb_scores = scalar(keys %tmp_scores);
				@scores = sort {$tmp_scores{$b} <=> $tmp_scores{$a}} (keys %tmp_scores);
				%tmp_scores=();
			}
			until($nb_scores < 1);
		}
	}
}


print "STEP 4 : WRITING THE OUTPUT FILE\n";
open(OUT, ">$output_file") || die("Can't open $output_file as OUT");
foreach my $acc (@acc)
{
	my @peptides = peptide_query($file, $acc, $acc{$acc}{dbtype});
	print OUT "BEGIN\n";
	print OUT "FILE=".$file."\n";
	print OUT "ACC=".$acc{$acc}{dbtype}."|".$acc."\n";
	print OUT "DESC=".$acc{$acc}{description}."\n";
	print OUT "ORG=".$acc{$acc}{organism}."\n";
	print OUT "DB=".$database."\n";
	# the score is an array, we must print the sum of the items in the array
#	print OUT "SCORE=".$acc{$acc}{score}."\n";
	my $tmp_score = 0;
	foreach my $myScore ($acc{$acc}{score})
	{
		$tmp_score += $myScore;
	}
	print OUT "SCORE=".$tmp_score."\n";
	print OUT "SEQUENCE=".$acc{$acc}{sequence}."\n";
	foreach my $peptide (@peptides)
	{
		print OUT "PEPTIDE=".$peptide."\n";
	}
	print OUT "END\n";
}
close(OUT);

print "END\n";

exit;

# search the file for the peptide sequences corresponding to the given accession number
# $file : the file to seach
# $acc : the accession number to search in the file
# dbtype : the type of database entry (gi, tr, sp)
# return an array of the sequences found in the file
sub peptide_query
{
	my ($file, $acc, $dbtype) = @_;
	my $peptide = "";
	my $nb_peptides = 0;
	my @peptides;
	open(INPUT, "$file") || die("Can't open $file as INPUT");
	# read the file until the accession number is found
	do {} until(($_=<INPUT>) =~ /^\>\/:$dbtype\|$acc/);
	# read the file while it concerns the accession number
	while(my $line = <INPUT>)
	{
		$line =~ s/\r?\n?$//;
		last if($line =~ m/^\>\/:/);
		# Query:   500 BVLDLAQDVVVR 511
		if($line =~ m/Query.*\d+\s(\w+)/) { push(@peptides, $1); }
		#elsif($line =~ m/Sbjct.*\d+\s(\w+)/) { $peptides[$#peptides] .= "#".$1; }
		#if($line =~ m/Query:/)
		#{
		#	$line =~ /.*\d+\s(\w+)/;
		#	push(@peptides, $1);
		#}
	}
	close (INPUT);
	return @peptides;
}

sub NEWFILE
{
	my $file = `mktemp -p \$(pwd)`;
	chomp($file);
	return $file;
}
